package rione.communication;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.worldmodel.EntityID;
import rione.io.BitArrayInputStream;
import rione.model.AdvancedWorldModel;
import rione.model.ConfigConstants.RadioChannel;
import rione.model.ConfigConstants.VoiceChannel;
import rione.model.Uniform;
import rione.util.BitUtil;

public class CommunicationUtil {

	public final int UINT_BIT_SIZE = 31;
	
	public final int FIRERYNESS_BIT_SIZE = Fieryness.values().length;
	public final int CIVILIAN_ID_BIT_SIZE = 31;
	public final int BURIEDNESS_BIT_SIZE = 31;
	public final int AT_UNIFORM_BIT_SIZE;
	public final int FB_UNIFORM_BIT_SIZE;
	public final int PF_UNIFORM_BIT_SIZE;
	public final int ROAD_UNIFORM_BIT_SIZE;
	public final int BUILDING_UNIFORM_BIT_SIZE;
	
	public final int ENTITY_ID_OFFSET;
	public final int ENTITY_ID_BIT_SIZE;
	
	public final int X_OFFSET;
	public final int X_BIT_SIZE;
	public final int Y_OFFSET;
	public final int Y_BIT_SIZE;
	
	public final int TIME_BIT_SIZE;
	
	public final int HP_BIT_SIZE;
	
	private final boolean noRadio;
	private final boolean noVoice;

	public CommunicationUtil(final AdvancedWorldModel world) {
		// Uniform関係
		final int AT_SIZE = world.getEntitiesOfType(StandardEntityURN.AMBULANCE_TEAM).size();
		this.AT_UNIFORM_BIT_SIZE = BitUtil.needBitSize(AT_SIZE);
		final int FB_SIZE = world.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE).size();
		this.FB_UNIFORM_BIT_SIZE = BitUtil.needBitSize(FB_SIZE);
		final int PF_SIZE = world.getEntitiesOfType(StandardEntityURN.POLICE_FORCE).size();
		this.PF_UNIFORM_BIT_SIZE = BitUtil.needBitSize(PF_SIZE);
		final int BUILDING_SIZE = world.getEntitiesOfType(StandardEntityURN.BUILDING).size();
		this.BUILDING_UNIFORM_BIT_SIZE = BitUtil.needBitSize(BUILDING_SIZE);
		final int ROAD_SIZE = world.getEntitiesOfType(StandardEntityURN.ROAD).size();
		this.ROAD_UNIFORM_BIT_SIZE = BitUtil.needBitSize(ROAD_SIZE);
		// Uniformが付けられないEntity用
		int maxID = 0;
		int minID = Integer.MAX_VALUE;
		for (StandardEntity se : world.getAllEntities()) {
			final int v = se.getID().getValue();
			maxID = Math.max(maxID, v);
			minID = Math.min(minID, v);
		}
		this.ENTITY_ID_OFFSET = minID;
		this.ENTITY_ID_BIT_SIZE = BitUtil.needBitSize(maxID - minID);
		//座標表現用
		final int MIN_X = world.getConfig().MIN_X;
		final int MIN_Y = world.getConfig().MIN_Y;
		final int MAX_X = world.getConfig().MAX_X;
		final int MAX_Y = world.getConfig().MAX_Y;
		this.X_OFFSET = MIN_X;
		this.X_BIT_SIZE = BitUtil.needBitSize(MAX_X - MIN_X);
		this.Y_OFFSET = MIN_Y;
		this.Y_BIT_SIZE = BitUtil.needBitSize(MAX_Y - MIN_Y);
		
		this.TIME_BIT_SIZE = BitUtil.needBitSize(world.getConfig().timestep);
		
		this.HP_BIT_SIZE = BitUtil.needBitSize(world.getConfig().hpMax / world.getConfig().hpPrecision);
		
		int totalRadioBandwidth = 0;
		for (RadioChannel c : world.getConfig().radioChannels.values()) {
			totalRadioBandwidth += c.bandwidth;
		}
		noRadio = (totalRadioBandwidth < 32) || (world.getConfig().subscribePlatoonSize <= 0);
		
		int totalVoiceSize = 0;
		int maxRange = 0;
		int maxNum = 0;
		for (VoiceChannel c : world.getConfig().voiceChannels.values()) {
			totalVoiceSize += c.size;
			maxRange = Math.max(maxRange, c.range);
			maxNum = Math.max(maxNum, c.maxNum);
		}
		noVoice = (totalVoiceSize < 32) || (maxNum <= 0) || (maxRange <= 50);
	}
	
	public boolean isNoRadio() {
		return noRadio;
	}
	public boolean isNoVoice() {
		return noVoice;
	}

	public void writeArea(MessageBitSection sec, final Area area, final Uniform uniform) {
		final int u = uniform.toUniform(area.getID());
		if (area instanceof Building) {
			sec.add(0, 1);
			sec.add(u, this.BUILDING_UNIFORM_BIT_SIZE);
		}
		else /*if (location instanceof Road)*/ {
			sec.add(1, 1);
			sec.add(u, this.ROAD_UNIFORM_BIT_SIZE);
		}
	}

	public EntityID readArea(BitArrayInputStream stream, final Uniform uniform) {
		int areaFlg = stream.readBit(1);
		if (areaFlg == 0) {
			return uniform.toID(StandardEntityURN.BUILDING, stream.readBit(this.BUILDING_UNIFORM_BIT_SIZE));
		}
		else /*if (areaFlg == 1)*/ {
			return uniform.toID(StandardEntityURN.ROAD, stream.readBit(this.ROAD_UNIFORM_BIT_SIZE));
		}
	}
}
