package rione.io;

import java.util.ArrayList;

/**
 * byte配列に情報をbit単位で書き込むためのクラス. 書き込む情報はint型に格納されている必要がある. 浮動小数および負の値は書き込めない.
 * 
 * @author com
 */
public class BitArrayOutputStream {

	/** 入力情報を一時的に格納するバッファ */
	private long buf;

	/** バッファに格納された情報のビットサイズ */
	private int offset;

	/** ビット列を格納する配列 */
	private final ArrayList<Byte> data;
	
	private int bitlength = 0;

	/**
	 * コンストラクタ
	 * 
	 */
	public BitArrayOutputStream() {
		this.buf = 0;
		this.offset = 0;
		this.data = new ArrayList<Byte>();
	}

	/**
	 * bit列の書き込まれたbyte配列を返す
	 * 
	 * @return bit列の書き込まれたbyte配列
	 */
	public byte[] getData() {
		flush();
		byte[] res = new byte[data.size()];
		int i = 0;
		for (byte b : data) {
			res[i++] = b;
		}
		return res;
	}

	/**
	 * bit列の書きこまれたbyteサイズを返す
	 * 
	 * @return bit列の書きこまれたbyteサイズ
	 */
	public int getDatasize() {
		return data.size();
	}
	
	public int getBitLength() {
		return bitlength;
	}

	/**
	 * ビット列とバッファを初期化する
	 */
	public void clear() {
		buf = 0;
		offset = 0;
		data.clear();
	}

	/**
	 * int型に格納された値から、下位nビット分を書き込む nは最大で32
	 * 
	 * @param val
	 *            書き込む値
	 * @param n
	 *            書き込むビット数
	 */
	public void writeBit(int val, int n) {
		bitlength += n;
		buf <<= n;
		buf += val & (0xFFFFFFFF >>> 32 - n);
		offset += n;
		while (offset >= 8) {
			offset -= 8;
			write((buf >>> offset) & 0xFF);
		}
	}

	/**
	 * long型に格納された値から,要素を1バイト書き込む
	 * 
	 * @param b
	 *            書き込む値
	 */
	private void write(long b) {
		data.add((byte) b);
	}

	/**
	 * バッファに格納されている情報を強制的に書き込む
	 */
	private void flush() {
		// バッファに何も書き込まれていない場合は何もしない
		if (offset == 0) {
			return;
		}
		write((buf << 8 - offset) & 0xFF);
		offset = 0;
	}

}
