package rione.model;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardPropertyURN;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import rione.communication.CivilianVoiceListener;
import rione.communication.MessageBitSection;
import rione.communication.Port;
import rione.communication.CommunicationUtil;
import rione.io.BitArrayInputStream;
import rione.model.AdvancedWorldModel;
import rione.util.BitUtil;
import rione.util.IntPropertySorter;
import rione.viewer.component.extension.BorderExtension;
import rione.viewer.component.extension.ExtensionMap;

final public class BuriedHumans implements Iterable<EntityID> {

	/** 埋まっているHuman */
	private Set<EntityID> humans = new HashSet<EntityID>();
	/** timestepまで生き残る/死亡している ことが確定しているHuman */
	private Set<EntityID> needless = new HashSet<EntityID>();
	private Set<EntityID> fromCommunication = new HashSet<EntityID>();
	
	private final AdvancedWorldModel world;
	
	private static final int MAX_N = 31;
	private static final int N_BIT_SIZE = BitUtil.needBitSize(MAX_N + 1);
	private static final int BURIEDNESS_SEND_MAX = 255;
	private static final int BURIEDNESS_BIT_SIZE = BitUtil.needBitSize(BURIEDNESS_SEND_MAX + 1);
	private final int MAX_HP;
	private final int MAX_HP_BIT;
	private final int MAX_DAMAGE;
	private final int MAX_DAMAGE_BIT;

	public BuriedHumans(AdvancedWorldModel w) {
		world = w;
		MAX_HP = world.getConfig().hpMax / world.getConfig().hpPrecision;
		MAX_HP_BIT = BitUtil.needBitSize(MAX_HP + 1);
		MAX_DAMAGE = world.getConfig().timestep / world.getConfig().damagePrecision;
		MAX_DAMAGE_BIT = BitUtil.needBitSize(MAX_DAMAGE + 1);
	}

	public void update(ChangeSet changed) {
		if (world.getTime() < world.getConfig().ignoreUntil) {
			return;
		}
		if (world.getTime() == world.getConfig().ignoreUntil) {
			for (StandardEntity se : world.getEntitiesOfType(AgentConstants.HUMANOIDS)) {
				Human human = (Human) se;
				if (shouldIn(human)) {
					humans.add(human.getID());
				}
			}
		}
		fromCommunication.removeAll(changed.getChangedEntities());
		// 諦め判定
		for (Iterator<EntityID> it = humans.iterator(); it.hasNext();) {
			Human human = (Human) world.getEntity(it.next());
			if (!shouldIn(human)) {
				it.remove();
			}
		}
		// 発見判定
		for (EntityID id : changed.getChangedEntities()) {
			StandardEntity se = world.getEntity(id);
			if (se instanceof Human) {
				Human human = (Human) se;
				if (shouldIn(human)) {
					humans.add(id);
				}
			}
		}
	}

	public CivilianVoiceListener createCivilianVoiceListener() {
		return new CivilianVoiceListener() {
			@Override
			public void hear(AKSpeak message) {
				// Voiceからの発見
				StandardEntity se = world.getEntity(message.getAgentID());
				if (se instanceof Human) {
					fromCommunication.add(se.getID());
					humans.add(se.getID());
				}
			}
		};
	}
	
	public boolean shouldIn(Human human) {
		if (!human.isPositionDefined()
				|| needless.contains(human.getID())) {
			return false;
		}
		StandardEntity position = human.getPosition(world);
		if (position.getStandardURN() == StandardEntityURN.REFUGE
				|| position.getStandardURN() == StandardEntityURN.ROAD
				|| position.getStandardURN() == StandardEntityURN.AMBULANCE_TEAM
				|| (human.isHPDefined() && human.getHP() == 0)) {
			needless.add(human.getID());
			return false;
		}
		if (fromCommunication.contains(human.getID())) {
			return true;
		}
		if (position instanceof Building) {
			if (human instanceof Civilian) {
				return true;
			}
			else if (human.isBuriednessDefined() && human.getBuriedness() == 0) {
				needless.add(human.getID());
				return false;
			}
			else {
				return true;
			}
		}
		else {
			if (human.isDamageDefined() && human.getDamage() == 0) {
				needless.add(human.getID());
			}
		}
		return false;
	}

	public Port createAddPort(final CommunicationUtil comUtil) {
		return new Port() {
			List<Human> addList;
			private boolean addSent;
			private ChangeSet changed;
			@Override
			public void init(ChangeSet changed) {
				this.changed = changed;
				addList = new ArrayList<Human>();
				for (EntityID id : humans) {
					Human human = (Human) world.getEntity(id);
					if (!fromCommunication.contains(id)) {
						addList.add(human);
					}
				}
				Collections.sort(addList, 
						new IntPropertySorter<StandardEntity>(
								StandardPropertyURN.BURIEDNESS));
				addSent = addList.isEmpty();
			}
			@Override
			public boolean hasNext() {
				return !addSent;
			}
			@Override
			public MessageBitSection next() {
				addSent = true;
				MessageBitSection sec = new MessageBitSection(10);
				final int n = Math.min(addList.size(), MAX_N);
				//救助対象の数n
				sec.add(n, N_BIT_SIZE);
				//n個の救助対象の情報
				for (int i = 0; i < n; i++) {
					Human human = addList.get(i);
					sec.add(human.getID().getValue(), comUtil.UINT_BIT_SIZE);
					sec.add(human.getPosition().getValue(), comUtil.UINT_BIT_SIZE);
					if (human.isBuriednessDefined()) {
						sec.add(Math.min(human.getBuriedness(), BURIEDNESS_SEND_MAX), BURIEDNESS_BIT_SIZE);
					}
					else {
						sec.add(BURIEDNESS_SEND_MAX, BURIEDNESS_BIT_SIZE);
					}
					int hp = (human.isHPDefined())
							? human.getHP() / world.getConfig().hpPrecision : MAX_HP;
					sec.add(hp, MAX_HP_BIT);
					int damage = (human.isDamageDefined())
							? Math.min(world.getConfig().timestep, human.getDamage() / world.getConfig().damagePrecision) : 0;
					sec.add(damage, MAX_DAMAGE_BIT);
				}
				return sec;
			}
			@Override
			public void read(EntityID sender, int time, BitArrayInputStream stream) {
				final int n = stream.readBit(N_BIT_SIZE);
				for (int i = 0; i < n; i++) {
					final EntityID id = new EntityID(stream.readBit(comUtil.UINT_BIT_SIZE));
					final EntityID posID = new EntityID(stream.readBit(comUtil.UINT_BIT_SIZE));
					final int buriedness = stream.readBit(BURIEDNESS_BIT_SIZE);
					final int hp = stream.readBit(MAX_HP_BIT);
					final int damage = stream.readBit(MAX_DAMAGE_BIT);
					try {
						Human human = (Human) world.getEntity(id);
						StandardEntity pos = world.getEntity(posID);
						if (!(pos instanceof Area || pos instanceof AmbulanceTeam)
								|| changed.getChangedEntities().contains(id)) {
							continue;
						}
						if (human == null) {
							human = world.addNewCivilian(id);
						}
						if (0 <= buriedness) {
							human.setBuriedness(buriedness);
						}
						else {
							human.setBuriedness(1);
						}
						if (0 <= hp && hp < MAX_HP) {
							human.setHP(hp * world.getConfig().hpPrecision);
						}
						else {
							human.setHP(world.getConfig().hpMax);
						}
						if (0 <= damage && damage < MAX_DAMAGE) {
							human.setDamage(damage);
						}
						if (pos != null && pos instanceof Building) {
							human.setPosition(posID);
						}
						fromCommunication.add(id);
						humans.add(id);
					}
					catch (ClassCastException e) {
						continue;
					}
				}
			}
		};
	}
	
	public Port createRemovePort(final CommunicationUtil comUtil) {
		return new Port() {
			List<Human> removeList;
			private boolean removeSent;
			@Override
			public void init(ChangeSet changed) {
				removeList = new ArrayList<Human>();
				for (EntityID id : changed.getChangedEntities()) {
					if (needless.contains(id)) {
						Human human = (Human) world.getEntity(id);
						removeList.add(human);
					}
				}
				removeSent = removeList.isEmpty();
			}
			@Override
			public boolean hasNext() {
				return !removeSent;
			}
			@Override
			public MessageBitSection next() {
				removeSent = true;
				MessageBitSection sec = new MessageBitSection(8);
				final int n = Math.min(removeList.size(), MAX_N);
				//救助対象の数n
				sec.add(n, N_BIT_SIZE);
				//n個の救助対象の情報
				for (int i = 0; i < n; i++) {
					Human human = removeList.get(i);
					sec.add(human.getID().getValue(), comUtil.UINT_BIT_SIZE);
				}
				return sec;
			}
			@Override
			public void read(EntityID sender, int time, BitArrayInputStream stream) {
				final int n = stream.readBit(N_BIT_SIZE);
				for (int i = 0; i < n; i++) {
					final EntityID id = new EntityID(stream.readBit(comUtil.UINT_BIT_SIZE));
					fromCommunication.add(id);
					humans.remove(id);
					needless.add(id);
					try {
						Human human = (Human) world.getEntity(id);
						if (human == null) {
							human = world.addNewCivilian(id);
						}
					}
					catch (ClassCastException e) {
					}
				}
			}
		};
	}
	
	public Port createNoComPort(final CommunicationUtil comUtil, boolean isAggregator, boolean isMessenger) {
		return null;
	}

	@Override
	public Iterator<EntityID> iterator() {
		return humans.iterator();
	}
	
	public boolean contains(EntityID id) {
		return humans.contains(id);
	}

	public void extension(ExtensionMap exMap) {
		for (StandardEntity se : world.getEntitiesOfType(StandardEntityURN.CIVILIAN)) {
			EntityID id = se.getID();
			exMap.put(id, new BorderExtension(exMap.get(id),Color.CYAN, 3.0f));
		}
		for (EntityID id : humans) {
			exMap.put(id, new BorderExtension(exMap.get(id), Color.ORANGE, 2.0f));
		}
	}

}
