package rione.model;

import java.awt.Color;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityConstants;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import rione.communication.Port;
import rione.communication.CommunicationUtil;
import rione.communication.MessageBitSection;
import rione.io.BitArrayInputStream;
import rione.util.BitUtil;
import rione.viewer.component.extension.BorderExtension;
import rione.viewer.component.extension.ExtensionMap;

@SuppressWarnings("serial")
public class BurningBuildings extends HashSet<Building> {
	
	/** そのサイクルで追加したEntity */
	private ArrayList<Building> addRecords = new ArrayList<Building>();
	/** そのサイクルで削除したEntity */
	private ArrayList<Building> removeRecords = new ArrayList<Building>();
	private AdvancedWorldModel world;
	
	public BurningBuildings(AdvancedWorldModel world) {
		this.world = world;
	}
	
	public void update(AdvancedWorldModel world, ChangeSet changeSet) {
		addRecords.clear();
		removeRecords.clear();
		for (EntityID e : changeSet.getChangedEntities()) {
			StandardEntity se = world.getEntity(e);
			if (se instanceof Building) {
				Building building = (Building) se;
				if (building.isOnFire()) {
					if (add(building)
							&& building.isFierynessDefined()
							&& building.isTemperatureDefined()) {
						addRecords.add(building);
					}
				}
				else if (remove(building)) {
					removeRecords.add(building);
				}
			}
		}
		// 初期発火を発見したら無理にでも伝える
		if (world.getTime() < world.getConfig().ignoreUntil + 8) {
			for (Building building : this) {
				if (building.isOnFire() && !addRecords.contains(building)) {
					addRecords.add(building);
				}
			}
		}
	}
	
	//TODO もっと細かく分ける
	private final int MAX_N = 31;
	private final int N_BIT_SIZE = BitUtil.needBitSize(MAX_N + 1);
	public Port createAddPort(final CommunicationUtil comUtil) {
		return new Port() {
			private boolean addSent;
			private ChangeSet changed;
			@Override
			public void init(ChangeSet changed) {
				this.changed = changed;
				addSent = addRecords.isEmpty();
			}
			@Override
			public boolean hasNext() {
				return !addSent;
			}
			@Override
			public MessageBitSection next() {
				addSent = true;
				MessageBitSection sec = new MessageBitSection(100);
				final int n = Math.min(addRecords.size(), MAX_N);
				//建物の数n
				sec.add(n, N_BIT_SIZE);
				//n個の建物の情報
				for (int i = 0; i < n; i++) {
					Building building = addRecords.get(i);
					int uniform = world.getUniform().toUniform(building.getID());
					sec.add(uniform, comUtil.BUILDING_UNIFORM_BIT_SIZE);
					sec.add(building.getFieryness(), comUtil.FIRERYNESS_BIT_SIZE);
					sec.add(building.getTemperature(), comUtil.UINT_BIT_SIZE);
				}
				return sec;
			}
			@Override
			public void read(EntityID sender, int time,
					BitArrayInputStream stream) {
				final int n = stream.readBit(N_BIT_SIZE);
				for (int i = 0; i < n; i++) {
					final int uniform = stream.readBit(comUtil.BUILDING_UNIFORM_BIT_SIZE);
					EntityID id = world.getUniform().toID(StandardEntityURN.BUILDING, uniform);
					final int fieryness = stream.readBit(comUtil.FIRERYNESS_BIT_SIZE);
					final int temperature = stream.readBit(comUtil.UINT_BIT_SIZE);
					if (id == null) continue;//ノイズでidがおかしい
					// 視界に入っていたら無視
					if (changed.getChangedEntities().contains(id))
						continue;
					Building building = (Building) world.getEntity(id);
					// ノイズのチェック
					if (0 <= fieryness
							&& fieryness < StandardEntityConstants.Fieryness.values().length) {
						building.setFieryness(fieryness);
					}
					else {
						// ノイズでおかしくなったらとりあえずBURNINGで
						building.setFieryness(StandardEntityConstants.Fieryness.BURNING.ordinal());
					}
					building.setTemperature(temperature);
					if (temperature > 0) {
						add(building);
					}
				}
			}
		};
	}
	
	public Port createRemovePort(final CommunicationUtil comUtil) {
		return new Port() {
			private boolean removeSent;
			@Override
			public void init(ChangeSet changed) {
				removeSent = removeRecords.isEmpty();
			}
			@Override
			public boolean hasNext() {
				return !removeSent;
			}
			@Override
			public MessageBitSection next() {
				removeSent = true;
				MessageBitSection sec = new MessageBitSection(50);
				final int n = Math.min(removeRecords.size(), MAX_N);
				//建物の数n
				sec.add(n, N_BIT_SIZE);
				//n個の建物の情報
				for (int i = 0; i < n; i++) {
					Building building = removeRecords.get(i);
					int uniform = world.getUniform().toUniform(building.getID());
					sec.add(uniform, comUtil.BUILDING_UNIFORM_BIT_SIZE);
				}
				return sec;
			}
			@Override
			public void read(EntityID sender, int time, BitArrayInputStream stream) {
				final int n = stream.readBit(N_BIT_SIZE);
				for (int i = 0; i < n; i++) {
					final int uniform = stream.readBit(comUtil.BUILDING_UNIFORM_BIT_SIZE);
					EntityID id = world.getUniform().toID(StandardEntityURN.BUILDING, uniform);
					if (id == null) continue;//ノイズでidがおかしい
					Building building = (Building) world.getEntity(id);
					if (building.isFierynessDefined()) {
						switch (building.getFierynessEnum()) {
						case UNBURNT:
						case BURNT_OUT:
							break;
						default:
							// とりあえずWATER_DAMAGEにしておく
							building.setFieryness(StandardEntityConstants.Fieryness.WATER_DAMAGE.ordinal());
						}
					}
					remove(building);
				}
			}
		};
	}
	
	public Port createNoComPort(final CommunicationUtil comUtil, final boolean isAggregator, final boolean isMessenger) {
		if (isAggregator) {
			return new NoComPort(comUtil, 180);
		}
		if (isMessenger) {
			return new NoComPort(comUtil, 80);
		}
		else {
			return new NoComPort(comUtil, 3);
		}
	}
		
	private class NoComPort implements Port {
		private boolean sent;
		private Set<EntityID> bringBuildings = new HashSet<EntityID>();
		final private CommunicationUtil comUtil;
		final private int remain;
		
		public NoComPort(CommunicationUtil comUtil, int remain) {
			this.comUtil = comUtil;
			this.remain = remain;
		}
		
		@Override
		public void init(ChangeSet changed) {
			for (Iterator<EntityID> it = bringBuildings.iterator(); it.hasNext();) {
				EntityID id = it.next();
				if (remain < world.getTime() - world.getTimestamp().getLastChangedTime(id)
						|| !world.getBurningBuildings().contains(id)) {
					it.remove();
				}
			}
			
			for (EntityID id : changed.getChangedEntities()) {
				StandardEntity se = world.getEntity(id);
				if (se.getStandardURN() == StandardEntityURN.BUILDING) {
					Building building = (Building) se;
					if (world.getBurningBuildings().contains(building)) {
						bringBuildings.add(building.getID());
					}
				}
			}
			sent = bringBuildings.isEmpty();
		}
		@Override
		public boolean hasNext() {
			return !sent;
		}
		@Override
		public MessageBitSection next() {
			sent = true;
			MessageBitSection sec = new MessageBitSection(50);
			final int n = Math.min(bringBuildings.size(), MAX_N);
			//建物の数n
			sec.add(n, N_BIT_SIZE);
			//n個の建物の情報
			List<EntityID> buildingsList = new ArrayList<EntityID>(bringBuildings);
			for (int i = 0; i < n; i++) {
				Building building = (Building) world.getEntity(buildingsList.get(i));
				int uniform = world.getUniform().toUniform(building.getID());
				sec.add(uniform, comUtil.BUILDING_UNIFORM_BIT_SIZE);
				sec.add(building.getFieryness(), comUtil.FIRERYNESS_BIT_SIZE);
				sec.add(building.getTemperature(), comUtil.UINT_BIT_SIZE);
				sec.add(world.getTimestamp().getLastChangedTime(building.getID()), comUtil.TIME_BIT_SIZE);
			}
			return sec;
		}
		@Override
		public void read(EntityID sender, int time, BitArrayInputStream stream) {
			final int n = stream.readBit(N_BIT_SIZE);
			for (int i = 0; i < n; i++) {
				final int uniform = stream.readBit(comUtil.BUILDING_UNIFORM_BIT_SIZE);
				EntityID id = world.getUniform().toID(StandardEntityURN.BUILDING, uniform);
				final int fieryness = stream.readBit(comUtil.FIRERYNESS_BIT_SIZE);
				final int temperature = stream.readBit(comUtil.UINT_BIT_SIZE);
				final int messageTime = stream.readBit(comUtil.TIME_BIT_SIZE);
				
				//ノイズでidがおかしい
				if (id == null) continue;
				//情報が古い
				if (messageTime < world.getTimestamp().getLastChangedTime(id)) continue;
				
				Building building = (Building) world.getEntity(id);
				
				// 持っていく情報に入れる
				bringBuildings.add(id);
				
				// ノイズのチェック
				if (0 <= fieryness
						&& fieryness < StandardEntityConstants.Fieryness.values().length) {
					building.setFieryness(fieryness);
				}
				else {
					// ノイズでおかしくなったらとりあえずBURNINGで
					building.setFieryness(StandardEntityConstants.Fieryness.BURNING.ordinal());
				}
				building.setTemperature(temperature);
				if (temperature > 0) {
					add(building);
				}
				world.getTimestamp().setLastChangedTime(id, messageTime);
			}
		}
	}
	
	public boolean contains(Building building) {
		return super.contains(building);
	}
	public boolean contains(EntityID id) {
		return super.contains(world.getEntity(id));
	}

	public void extensions(ExtensionMap exMap) {
		exMap.putExtensionToAll(this, 
				new BorderExtension(Color.RED, 3.0f).decorator());
	}
}
