package rione.model.route.pov;

import java.util.Map;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Hydrant;
import rescuecore2.standard.entities.Road;
import rescuecore2.worldmodel.EntityID;
import rione.model.AdvancedWorldModel;
import rione.model.route.pov.graph.AreaNode;
import rione.model.route.pov.graph.EdgeNode;
import rione.model.route.pov.graph.PassableDictionary;
import rione.model.route.pov.graph.PointNode;

class CostFunctionFactory {
	
	/**
	 * 瓦礫を考慮し最短時間で到達できる経路を求めるときに使う．
	 * @param world
	 * @return
	 */
	static CostFunction normal(final AdvancedWorldModel world, final PassableDictionary passableDic) {
		return new CostFunction() {
			@Override
			public double cost(final PointNode from, final PointNode to) {
				double distance = from.distance(to);
				AreaNode areaNode;
				EdgeNode edgeNode;
				if (from instanceof AreaNode) {
					areaNode = (AreaNode) from;
					edgeNode = (EdgeNode) to;
				}
				else /* if (to instanceof AreaNode) */ {
					edgeNode = (EdgeNode) from;
					areaNode = (AreaNode) to;
				}
				Area area = areaNode.getBelong();
				if (area instanceof Building) {
					// 1サイクルでうまく抜けるとノーダメージで通れるが…
					if (((Building) area).isOnFire()) {
						return distance * 1000.0;
					}
				}
				if (area instanceof Road || area instanceof Hydrant) {
					switch (passableDic.getPassableLevel(areaNode, edgeNode)) {
					case SURE_PASSABLE:
					case COMMUNICATION_PASSABLE:
						return distance;				
					case LOGICAL_PASSABLE:
						return distance * 10.0;
					case UNKNOWN:
						return distance;
					case UNPASSABLE:
						return distance * 100000000.0;
					}
				}
				return distance;
			}
		};
	}
	
	/**
	 * UNPASSABLEに対して厳格に処理する．
	 * SureReachableが含まれた集合に対してのmoveでなければ計算が間に合わない恐れがある．
	 * @param world
	 * @return
	 */
	static CostFunction strict(final AdvancedWorldModel world, final PassableDictionary passableDic) {
		return new CostFunction() {
			@Override
			public double cost(final PointNode from, final PointNode to) {
				double distance = from.distance(to);
				AreaNode areaNode;
				EdgeNode edgeNode;
				if (from instanceof AreaNode) {
					areaNode = (AreaNode) from;
					edgeNode = (EdgeNode) to;
				}
				else /* if (to instanceof AreaNode) */ {
					edgeNode = (EdgeNode) from;
					areaNode = (AreaNode) to;
				}
				Area area = areaNode.getBelong();
				if (area instanceof Building) {
					// 1サイクルでうまく抜けるとノーダメージで通れるが…
					if (((Building) area).isOnFire()) {
						return distance * 100000.0;
					}
				}
				if (area instanceof Road || area instanceof Hydrant) {
					switch (passableDic.getPassableLevel(areaNode, edgeNode)) {
					case SURE_PASSABLE:
					case COMMUNICATION_PASSABLE:
						return distance;				
					case LOGICAL_PASSABLE:
						return distance * 10.0;
					case UNKNOWN:
						return distance * 2;
					case UNPASSABLE:
						return Double.POSITIVE_INFINITY;
					}
				}
				return distance;
			}
		};
	}
	
	/**
	 * UNPASSABLEのコストをnormalより若干下げる．
	 * @param world
	 * @return
	 */
	static CostFunction search(final AdvancedWorldModel world, final PassableDictionary passableDic) {
		return new CostFunction() {
			@Override
			public double cost(final PointNode from, final PointNode to) {
				double distance = from.distance(to);
				AreaNode areaNode;
				EdgeNode edgeNode;
				if (from instanceof AreaNode) {
					areaNode = (AreaNode) from;
					edgeNode = (EdgeNode) to;
				}
				else /* if (to instanceof AreaNode) */ {
					edgeNode = (EdgeNode) from;
					areaNode = (AreaNode) to;
				}
				Area area = areaNode.getBelong();
				if (area instanceof Building) {
					// 1サイクルでうまく抜けるとノーダメージで通れるが…
					if (((Building) area).isOnFire()) {
						return distance * 1000.0;
					}
				}
				if (area instanceof Road || area instanceof Hydrant) {
					switch (passableDic.getPassableLevel(areaNode, edgeNode)) {
					case SURE_PASSABLE:
						return distance * 1.5;
					case COMMUNICATION_PASSABLE:
						return distance;		
					case LOGICAL_PASSABLE:
						return distance * 10.0;
					case UNKNOWN:
						return distance * 0.75;
					case UNPASSABLE:
						return distance * 1000000.0;
					}
				}
				return distance;
			}
		};
	}
	
	/**
	 * 目的地周辺(r < ExtinguishDistance)ならUNPASSABLEを無視
	 * @param world
	 * @return
	 */
	static CostFunction fb(final AdvancedWorldModel world, final PassableDictionary passableDic, final Building dest) {
		return new CostFunction() {
			@Override
			public double cost(final PointNode from, final PointNode to) {
				double distance = from.distance(to);
				AreaNode areaNode;
				EdgeNode edgeNode;
				if (from instanceof AreaNode) {
					areaNode = (AreaNode) from;
					edgeNode = (EdgeNode) to;
				}
				else /* if (to instanceof AreaNode) */ {
					edgeNode = (EdgeNode) from;
					areaNode = (AreaNode) to;
				}
				Area area = areaNode.getBelong();
				if (area instanceof Building) {
					// 1サイクルでうまく抜けるとノーダメージで通れるが…
					if (((Building) area).isOnFire()) {
						return distance * 2.0;
					}
				}
				if (to.distance(dest.getX(), dest.getY()) < world.getConfig().extinguishableDistance) {
					return distance;
				}
				if (area instanceof Road || area instanceof Hydrant) {
					switch (passableDic.getPassableLevel(areaNode, edgeNode)) {
					case SURE_PASSABLE:
					case COMMUNICATION_PASSABLE:
						return distance;				
					case LOGICAL_PASSABLE:
						return distance * 10.0;
					case UNKNOWN:
						return distance;
					case UNPASSABLE:
						return Math.max(
								distance * 1000.0 * to.distance(dest.getX(), dest.getY()) / world.getConfig().extinguishableDistance,
								1);
					}
				}
				return distance;
			}
		};
	}

	/**
	 * 瓦礫を無視した最短経路を求める<br>
	 * PFが最短経路を開けるときに使う．
	 * @param world
	 * @return
	 */
	static CostFunction pf(final AdvancedWorldModel world) {
		return new CostFunction() {
			@Override
			public double cost(PointNode from, PointNode to) {
				return from.distance(to);
			}
		};
	}
	
	/**
	 * Buriednessを考慮する
	 * @param world
	 * @param passableDic
	 * @return
	 */
	static CostFunction at(
			final AdvancedWorldModel world,
			final PassableDictionary passableDic,
			final Map<EntityID, Double> minStaticCost) {
		return new CostFunction() {
			@Override
			public double cost(final PointNode from, final PointNode to) {
				double distance = from.distance(to);
				AreaNode areaNode;
				EdgeNode edgeNode;
				if (from instanceof AreaNode) {
					areaNode = (AreaNode) from;
					edgeNode = (EdgeNode) to;
				}
				else /* if (to instanceof AreaNode) */ {
					edgeNode = (EdgeNode) from;
					areaNode = (AreaNode) to;
				}
				Area area = areaNode.getBelong();
				if (area instanceof Building) {
					// 1サイクルでうまく抜けるとノーダメージで通れるが…
					if (((Building) area).isOnFire()) {
						return distance * 100000.0;
					}
				}
				if (area instanceof Road || area instanceof Hydrant) {
					switch (passableDic.getPassableLevel(areaNode, edgeNode)) {
					case SURE_PASSABLE:
					case COMMUNICATION_PASSABLE:
						return distance;				
					case LOGICAL_PASSABLE:
						return distance * 10.0;
					case UNKNOWN:
						return distance * 2;
					case UNPASSABLE:
						return distance * 100000000.0;
					}
				}
				Double staticCost = minStaticCost.get(area.getID());
				if (staticCost != null) {
					distance += staticCost;
				}
				return distance;
			}
		};
	}
}
