package rione.model.route.pov;

import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;
import rione.model.AdvancedWorldModel;
import rione.model.route.pov.graph.AreaNode;
import rione.model.route.pov.graph.EdgeNode;
import rione.model.route.pov.graph.PassableDictionary;
import rione.model.route.pov.graph.PassableDictionary.PassableLevel;
import rione.model.route.pov.graph.PointOfVisivility;
import rione.model.route.pov.reachable.BFSReachableArea;
import rione.model.route.pov.reachable.UFTReachableArea;

/**
 * Point of View 形式の経路探索．<br>
 * rione.model.route.povの窓口(Facade)でもある．<br>
 * 新しい経路探索を作りたくなったらコイツを置き換えるように作れば良い．
 * @author utisam
 *
 */
public class POVRouter {
	
	private AdvancedWorldModel world;
	private final Human me;
	
	private final PassableDictionary passableDic;
	private BFSReachableArea bfsReachable;
	private final UFTReachableArea uftReachable;
	private final PointOfVisivility pov;
	private final POVSearch search;
	
	public POVSearch search() {
		return search;
	}

	private final CostFunction normalFunc;
	private final CostFunction pfFunc;
	private final CostFunction searchFunc;


	public PassableDictionary getPassableDic() {
		return passableDic;
	}

	public POVRouter(Human me, AdvancedWorldModel world) {
		this.world = world;
		this.me = me;
		
		// 各サブモジュールの作成
		pov = new PointOfVisivility(world);
		passableDic = new PassableDictionary(world);
		bfsReachable = new BFSReachableArea(world);
		uftReachable = new UFTReachableArea(world);
	
		// 経路探索を行う時間を決める
		final int routeThinkTime = world.getConfig().thinkTime * 2 / 3;
		search = new POVSearch(routeThinkTime, pov);
		
		// CostFunctionの生成
		normalFunc = CostFunctionFactory.normal(world, passableDic);
		pfFunc = CostFunctionFactory.pf(world);
		searchFunc = CostFunctionFactory.search(world, passableDic);
	}
	
	public void update(final EntityID pos, final Set<EntityID> visibleEntitiesID) {
		Set<EdgeNode> newPassables = passableDic.update(pos, this, visibleEntitiesID);
		if (bfsReachable != null) {
			try {
				bfsReachable.update(pov.get(pos), world);
			}
			catch (IllegalStateException e) {
				// 遅くなるならやめる
				System.out.println("off: BFS Reachable update");
				bfsReachable = null;
			}	
		}
		if (uftReachable != null) {
			/*Pair<Integer, Integer> location = me.getLocation(world);
			uftReachable.uftMoveError(location);*/
			uftReachable.update(world, newPassables);
		}
		
	}

	/**
	 * A*探索 単一の目的地に移動する際に用いる
	 * 
	 * @param origin
	 * @param destinations
	 * @param costFunc
	 * @param world
	 * @return
	 */
	public List<EntityID> getAStar(Human me, Area destination, CostFunction costFunc) {
		POVPath result = search.getAStarPath(me, destination, costFunc);
		if (result == null) {
			return Collections.singletonList(me.getPosition());
		}
		return result.getRoute();
	}
	public List<EntityID> getAStar(Area from, Area destination, CostFunction costFunc) {
		POVPath result = search.getAStarPath(from, destination, costFunc);
		if (result == null) {
			return Collections.singletonList(from.getID());
		}
		return result.getRoute();
	}
	
	public List<EntityID> getMultiDest(Area origin, Collection<? extends StandardEntity> destinations, CostFunction costFunc) {
		return getMultiAStar(origin, destinations, costFunc);
		//return getDijkstra(origin, new HashSet<StandardEntity>(destinations), costFunc);
	}
	/**
	 * ダイクストラ探索 複数の目的地に移動する際に用いる
	 * 
	 * @param origin
	 * @param destinations
	 * @param costFunc
	 * @param world
	 * @return
	 */
	public List<EntityID> getDijkstra(Area origin, Set<StandardEntity> destinations, CostFunction costFunc) {
		POVPath result = search.getDijkstraPath(origin, destinations, costFunc);
		if (result == null) {
			return Collections.singletonList(origin.getID());
		}
		return result.getRoute();
	}
	/**
	 * MultiAStar探索 複数の目的地に移動する際に用いる
	 * 効果は不明
	 * @param origin
	 * @param destinations
	 * @param costFunc
	 * @param world
	 * @return
	 */
	public List<EntityID> getMultiAStar(Area origin, Collection<? extends StandardEntity> destinations, CostFunction costFunc) {
		POVPath result = search.getMultiAStarPath(origin, destinations, costFunc);
		if (result == null) {
			return Collections.singletonList(origin.getID());
		}
		return result.getReverseRoute();
	}
	
	public CostFunction getNormalCostFunction() {
		return normalFunc;
	}
	public CostFunction getPfCostFunction() {
		return pfFunc;
	}
	public CostFunction getSearchCostFunction() {
		return searchFunc;
	}
	public CostFunction getFbCostFunction(Building dest) {
		return CostFunctionFactory.fb(world, passableDic, dest);
	}
	public CostFunction getAtCostFunction(final Map<EntityID, Double> minStaticCost) {
		return CostFunctionFactory.at(world, passableDic, minStaticCost);
	}

	public PointOfVisivility getPOV() {
		return pov;
	}
	
	/**
	 * *_PASSABLE系の辺のみで到達可能
	 * @param area
	 * @return
	 */
	public boolean isSureReachable(Area area) {
		return isSureReachable(area.getID());
	}
	
	/**
	 * *_PASSABLE系の辺のみで到達可能
	 * @param area
	 * @return
	 */
	public boolean isSureReachable(EntityID id) {
		if (bfsReachable != null) {
			return bfsReachable.isSureReachable(id);
		}
		if (uftReachable != null) {
			return uftReachable.isSureReachable(me.getPosition(), id);
		}
		return false;
	}
	
	/**
	 * UNPASSABLE以外で到達可能(UNKNOWNを含む)．
	 * だたし，マップが広いと常にfalseになります．
	 * @param area
	 * @return
	 */
	public boolean isReachable(Area area) {
		return isReachable(area.getID());
	}
	/**
	 * UNPASSABLE以外で到達可能(UNKNOWNを含む)．
	 * だたし，マップが広いと常にfalseになります．
	 * @param area
	 * @return
	 */
	public boolean isReachable(EntityID id) {
		if (bfsReachable != null) {
			return bfsReachable.isReachable(id);
		}
		return false;
	}

	/**
	 * エリアから各Edgeへの線分のPassableLevelを数えます
	 * @param areaID
	 * @param level
	 * @return エリアから各Edgeへの線分がlevelと等しい数
	 */
	public int getPassableLevelCount(EntityID areaID, PassableLevel level) {
		final AreaNode anode = pov.get(areaID);
		if (anode == null) return 0;
		int count = 0;
		for (EdgeNode p : anode.getNeighbours()) {
			if (getPassableDic().getPassableLevel(anode, p) == level) {
				count++;
			}
		}
		return count;
	}
	
	public POVPath amend(Human me, POVPath old) {
		AreaNode currentNode = pov.get(me.getPosition());
		for (POVPath current = old; current.getPrevious() != null; current = current.getPrevious()) {
			if (current.getPoint().equals(currentNode)) {
				System.out.println("amend old route");
				current.setPrevious(null);
				return old;
			}
    	}
		// ルートから外れた時
		//TODO テスト不十分
		System.out.println("Out of old route");
		return search.getAStarPath(
				me,
				(Area) world.getEntity(old.getRoute().get(0)),
				getNormalCostFunction()).add(old);
	}

}
