package rione.model.route.pov.graph;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Shape;
import java.awt.geom.Line2D;
import java.awt.geom.PathIterator;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;
import rione.geom.ExpandApexes;
import rione.model.AdvancedWorldModel;
import rione.model.AgentConstants;
import rione.model.route.pov.POVPath;
import rione.model.route.pov.POVRouter;
import rione.util.SetPair;
import rione.viewer.component.TransformedGraphics;

/**
 * 到達可能かを保存しておく辞書
 * @author utisam
 *
 */
public class PassableDictionary {

	public enum PassableLevel {
		// 上のほうが優先度が高い(上書きされない)
		/** 自分の目で確認してPassable */
		SURE_PASSABLE, 
		/** 通信で得たPassable */
		COMMUNICATION_PASSABLE,
		/** 外部の情報から考えて論理的にPassable */
		LOGICAL_PASSABLE,
		/** Unpassable */
		UNPASSABLE,
		/** Unknown */
		UNKNOWN; /* NULL */
		public boolean isPassable() {
			return this.ordinal() < UNPASSABLE.ordinal();
		}
	}
	
	final private HashMap<SetPair<PointNode, PointNode>, PassableLevel> passableMap;
	final private AdvancedWorldModel world;
	
	@SuppressWarnings("serial")
	public PassableDictionary(AdvancedWorldModel world) {
		this.world = world;
		// (初期容量 * 負荷係数)を超えたときにMapの確保領域を増やす
		// TODO 初期容量と負荷係数を上手く設定してパフォーマンスをチューニング
		int areaSize = world.getEntitiesOfType(AgentConstants.AREAS).size();
		passableMap = new HashMap<SetPair<PointNode, PointNode>, PassableLevel>(areaSize * 2) {
			@Override
			public PassableLevel get(final Object o) {
				final PassableLevel value = super.get(o);
				return (value == null) ? PassableLevel.UNKNOWN : value;
			}
			@Override
			public PassableLevel put(final SetPair<PointNode, PointNode> key, final PassableLevel value) {
				final PassableLevel preValue = super.put(key, value);
				return (preValue == null) ? PassableLevel.UNKNOWN : preValue;
			}
		};
	}
	
	/**
	 * 
	 * @param pos
	 * @param router
	 * @param visibleEntitiesID
	 * @return 新しくPassableになったEdgeNode
	 */
	public Set<EdgeNode> update(final EntityID pos, final POVRouter router, final Set<EntityID> visibleEntitiesID) {
		final Set<EdgeNode> result = new HashSet<EdgeNode>();
		final PointOfVisivility pov = router.getPOV();
		
		result.addAll(lineCrossJudge(visibleEntitiesID, pov));
		
		// 斥力で動いてしまうので開始時は行わない
		if (world.getTime() > world.getConfig().ignoreUntil) {
			result.addAll(preOwnRoutePath((Area) world.getEntity(pos), router));
			// エージェントをあまり送らない通信システムでは完全に役立たず
			result.addAll(preAgentsRoutePath(router));
		}
		return Collections.unmodifiableSet(result);
	}

	private List<Pair<EntityID, EntityID>> preAgentArea = new ArrayList<Pair<EntityID, EntityID>>();
	/**
	 * 前サイクルのエージェントの位置から判断する<br>
	 * LogicalPassableを判定する
	 * @param router
	 * @return 
	 */
	private Set<EdgeNode> preAgentsRoutePath(final POVRouter router) {
		final Set<EdgeNode> result = new HashSet<EdgeNode>();
		for (Pair<EntityID, EntityID> agentPreArea : preAgentArea) {
			Human human = (Human) world.getEntity(agentPreArea.first());
			if (!human.isPositionDefined()) continue;
			StandardEntity currentArea = human.getPosition(world);
			StandardEntity pastArea = world.getEntity(agentPreArea.second());
			if (currentArea instanceof Area && pastArea != null) {
				POVPath path = router.search().getAStarPath((Area) pastArea, (Area) currentArea,
						router.getPfCostFunction());
				if (path == null)  {
					continue;
				}
				for (POVPath current = path; current.getPrevious() != null; current = current.getPrevious()) {
//					System.out.println("LogicalPassable");
					final SetPair<PointNode, PointNode> key = new SetPair<PointNode, PointNode>(current.getPoint(), current.getPrevious().getPoint());
					final PassableLevel currentLevel = passableMap.get(key);
					if (currentLevel.ordinal() < PassableLevel.LOGICAL_PASSABLE.ordinal())
						continue;
					passableMap.put(key, PassableLevel.LOGICAL_PASSABLE);
					if (!currentLevel.isPassable()) {
						if (current.getPoint() instanceof EdgeNode) {
							result.add((EdgeNode) current.getPoint());
						}
						else if (current.getPrevious().getPoint() instanceof EdgeNode) {
							result.add((EdgeNode) current.getPrevious().getPoint());
						}
					}
		    	}
			}
		}
		// 次のサイクルのために位置を保存
		preAgentArea.clear();
		for (StandardEntity se : world.getEntitiesOfType(AgentConstants.HUMANOIDS)) {
			Human human = (Human) se;
			if (human.isPositionDefined()) {
				StandardEntity currentArea = human.getPosition(world);
				if (currentArea instanceof Area) { 
					preAgentArea.add(new Pair<EntityID, EntityID>(human.getID(), currentArea.getID()));
				}
			}
		}
		return result;
	}

	/** 前サイクルの自分の位置 */
	private EntityID preArea = null;
	/**
	 * 前サイクルの自分の位置と現在位置から判断 
	 * @param currentOwnArea
	 * @param router
	 * @return 
	 */
	private Set<EdgeNode> preOwnRoutePath(final Area currentOwnArea, final POVRouter router) {
		Set<EdgeNode> result = new HashSet<EdgeNode>();
		if (preArea != null) {
			final POVPath path = router.search().getAStarPath((Area) world.getEntity(preArea), currentOwnArea,
					router.getPfCostFunction());
			if (path != null)  {
				for (POVPath current = path; current.getPrevious() != null; current = current.getPrevious()) {
					final SetPair<PointNode, PointNode> key = new SetPair<PointNode, PointNode>(current.getPoint(), current.getPrevious().getPoint());
					final PassableLevel value = passableMap.get(key);
					if (value == PassableLevel.UNKNOWN ||
							value.ordinal() < PassableLevel.LOGICAL_PASSABLE.ordinal())
						continue;
					passableMap.put(key, PassableLevel.LOGICAL_PASSABLE);
					if (!value.isPassable()) {
						if (current.getPoint() instanceof EdgeNode) {
							result.add((EdgeNode) current.getPoint());
						}
						else if (current.getPrevious().getPoint() instanceof EdgeNode) {
							result.add((EdgeNode) current.getPrevious().getPoint());
						}
					}
		    	}
			}
		}
		preArea = currentOwnArea.getID();
		return result;
	}

	/**
	 * 目で見たRoadの判断<br>
	 * SURE_PASSABLEかUNPASSABLEのどちらかになる
	 * @param visibleEntitiesID
	 * @param pov
	 * @return 新しくPassableになったEdgeNode
	 */
	private Set<EdgeNode> lineCrossJudge(Set<EntityID> visibleEntitiesID, PointOfVisivility pov) {
		Set<EdgeNode> result = new HashSet<EdgeNode>();
		for (EntityID id : visibleEntitiesID) {
			if (world.getEntity(id) instanceof Road) {
				AreaNode areaNode = pov.get(id);
				if(areaNode != null){
					for (EdgeNode edgeNode : areaNode.getNeighbours()) {
						//if (getPassableLevel(areaNode, edgeNode) == PassableLevel.SURE_PASSABLE)
						//	continue;
						// 目で見た判断なのでほぼ無条件にput
						if(edgeNode != null){
							PassableLevel level = isPassableLine(areaNode, edgeNode);
							if (level != PassableLevel.UNKNOWN && level != null) {
								PassableLevel preValue = passableMap.put(new SetPair<PointNode, PointNode>(areaNode, edgeNode), level);
								if ((!preValue.isPassable()) && level.isPassable()) {
									result.add(edgeNode);
								}
							}
						}
					}
				}
			}
		}
		return result;
	}

	private PassableLevel isPassableLine(final AreaNode areaNode, final EdgeNode edgeNode) {
		Area area = areaNode.getBelong();
		if (!area.isBlockadesDefined()) return PassableLevel.UNKNOWN;
		for (EntityID id : area.getBlockades()) {
			final Blockade block = (Blockade) world.getEntity(id);
			final Shape shape = ExpandApexes.expandApexes(block, 800);
			if (shape == null) return PassableLevel.UNKNOWN;
			// 2点のどちらかを含んでいたら UNPASSABLE
			if (shape.contains(areaNode.getX(), areaNode.getY())
					|| shape.contains(edgeNode.getX(), edgeNode.getY())) {
				return PassableLevel.UNPASSABLE;
			}
			final Line2D pathLine = new Line2D.Double(
					areaNode.getX(), areaNode.getY(),
					edgeNode.getX(), edgeNode.getY());
			PathIterator it = shape.getPathIterator(null);
			double[] c = new double[6];
			double x = 0.0, y = 0.0;
			while (!it.isDone()) {
				switch (it.currentSegment(c)) {
				case PathIterator.SEG_MOVETO:
					x = c[0];
					y = c[1];
					break;
				case PathIterator.SEG_LINETO:
					if (pathLine.intersectsLine(x, y, c[0], c[1])) {
						return PassableLevel.UNPASSABLE;
					}
					x = c[0];
					y = c[1];
					break;
				case PathIterator.SEG_QUADTO:
				case PathIterator.SEG_CUBICTO:
				case PathIterator.SEG_CLOSE:
					break;
				}
				it.next();
			}
		}
		return PassableLevel.SURE_PASSABLE;
	}

	public PassableLevel getPassableLevel(final PointNode a, final PointNode b) {
		// RootNodeは個別に計算する．
		if (a instanceof RootNode) {
			return isPassableLine((AreaNode) a, (EdgeNode) b);
		}
		if (b instanceof RootNode) {
			return isPassableLine((AreaNode) b, (EdgeNode) a);
		}
		// BuildingにはBlockadeは存在しない
		if ((a instanceof AreaNode && ((AreaNode) a).getBelong() instanceof Building)
				|| (b instanceof AreaNode && ((AreaNode) b).getBelong() instanceof Building)) {
			return PassableLevel.SURE_PASSABLE;
		}
		SetPair<PointNode, PointNode> key = new SetPair<PointNode, PointNode>(a, b);
		PassableLevel level = passableMap.get(key);
		return level;
	}

	public void customRender(TransformedGraphics gt) {
		gt.setStroke(new BasicStroke());
		gt.setThickness(3);
		final PointOfVisivility pov = world.getRouter().getPOV();
		for (AreaNode n : pov.getAreaNodes()) {
			for (EdgeNode e : n.getNeighbours()) {
				switch(this.getPassableLevel(n, e)) {
				case SURE_PASSABLE:
					gt.setColor(Color.CYAN);
					break;
				case COMMUNICATION_PASSABLE:
					gt.setColor(Color.YELLOW);
					break;
				case LOGICAL_PASSABLE:
					gt.setColor(Color.GREEN);
					break;
				case UNPASSABLE:
					gt.setColor(Color.RED);
					break;
				case UNKNOWN:
					gt.setColor(Color.WHITE);
					break;
				}
				gt.drawLine(n.getX(), n.getY(), e.getX(), e.getY());					
			}
		}
	}
}
